################################################################################
##                                                                            ##
## ------------ CRAZY RANKS ------------------------------------------------- ##
##                                                                            ##
## DEVELOPED BY: .craziesT                                                    
##                                                                            
## DESCRIPTION: A fully-featured ranking system. Fully extendable and         
## customizable. Requires no more than Eventscripts 2.0. For more information 
## please read the included config/help file. For any feature requests /      
## bug reports please contact me on the forum thread. Thanks, craziesT.       
##                                                                            
## VERSION: 1.0.2 - Top System Incorporated (not tested yet!)                 
## VERSION: 1.0.3 - Multilingual support added (see languages.ini)            
## VERSION: 1.0.4 - Added Config file for easier editing (see config.cfg)     
## VERSION: 1.0.5 - Added support for topx commands                           
## VERSION: 1.0.6 - Addded weapon multipliers (see config.cfg)                
## VERSION: 1.0.7 - Added french translation (thanks to L'In20Cible!)         
## VERSION: 1.0.8 - Added danish translation (thanks to TheCheeTaH!)          
## VERSION: 1.0.9 - Added portuguese translation (thanks to Glazz)            
## VERSION: 1.1a - Added weaponme command system                              
## VERSION: 1.1b - Fixed small bug (changed the translations file to UTF-8)   
## VERSION: 1.2a - Added russian traslation (thanks to chameleon82)           
## VERSION: 1.3a - Fixed bug with top menu not showing player names           
## VERSION: 1.3b - Added command control and sound control for achievements   
##
## NEW CHANGE LIST:
## Version 1.4c:
##  > FIX: The translations file should not work properly when the ranks
##    command is used by a player.
##
## Version 1.4b:
##  > FIX: Fixed the do_bots_count command, bots should not be ranked now.
##  > FIX: The top menu should function normally now
##  > FIX: The external modules should work perfectly now
##  > FIX: The players game time is now fixed!
##  > ADDED: An export function which allows you to export the ranks data and
##    in the future a web display will be included
##  > CHANGE: A new in-script settings class for easier settings handling
##    (credit to DanielB for writing the class)
##  > CHANGE: Rewrote the player class
##
## Version 1.4a:
##  > CHANGE: The display of top menu does not show the 8. - and the 9. -
##    lines anymore,
##  > FIX: The top menu does not show negative positions anymore,
##  > ADDED: A console message output when the player joins, showing him all
##    the commands that may be used and notifies them of this,
##  > ADDED: An option to update the database with the new system automatically
##    if the database existed before the update,
##  > ADDED: Option to store players by steamid, ip address or by name, changed
##    most of the script so it supports the change,
##  > ADDED: A message when a player dies to see how many points the attacker
##    got for killing them,
##  > ADDED: New module to allows players to see their ip, steamid
##  > REPLACED: Function botCheck(userid) with player.getUniqueId()
##                                                                            ##
##                                                                            ##
##                                                                            ##
################################################################################

## TO-DO LIST :

## Player Achievements (need to incorporate)
## Deduction System

''' Imports '''
from __future__ import division
import es, cmdlib, usermsg, playerlib, popuplib, langlib, cfglib
import sqlite3, time, psyco
from path import path

psyco.full()
''' Addon Information '''
info = es.AddonInfo()
info.name = "Crazy Ranks"
info.basename = "crazyranking"
info.author = ".craziesT"
info.description = "A fully-featured ranking system."

''' Global Variables '''
__version__ = "1.4c"
addon_path = es.getAddonPath("crazyranking")
public_variable = "%s_version" % info.basename
es.ServerVar(public_variable, __version__, 'Crazy Ranks Version').makepublic()
text = langlib.Strings(addon_path + '/config/languages.ini')

class Settings(object):
    config = cfglib.AddonCFG(addon_path + "/config/config.cfg")
    
    config.text("**********************************************************")
    config.text("** Crazy Ranks Plugin ********* Developed by: .craziesT **")
    config.text("**********************************************************")
    
    config.text(" ")
    config.text("Main Configuration File")
    config.text(" ")
    config.text("Edit the settings below to fit the script to your needs,")
    config.text("if you would like to request a new feature please post in")
    config.text("in the scripts forum thread or for any support also head to the")
    config.text("forums. You can also reach me at my steam account: craziest994")
    config.text(" ")
    config.text("Load the plugin with: es_load crazyranking")
    config.text(" ")
    config.text("You can translate the plugin to to other languages by editing the")
    config.text("languages.ini file. If you find ti confusing, you may request a")
    config.text("translation on the forums.")
    config.text(" ")
    config.text("Special thanks to L'In20Cible for the french translation. :)")
    config.text("Special thanks to TheCheeTaH for the danish translation. :)")
    config.text("Special thanks to Glazz for the portuguese translation. :)")
    config.text("Special thanks to chameleon82 for the russian translation. :)")
    config.text(" ")
    
    config.text('*****************************')
    config.text('Main Settings')
    config.text('*****************************')
    
    _points_system      = config.cvar("crazyranks_points_system",            1,         "Rank player by points or by kills (1 = by points, 0 = by kills) <points is recommended>"), int
    _multiplier         = config.cvar("crazyranks_points_multiplier",        10,        "The multiplier for the points gain by formula: (victim_points/attacker_points) * multiplier"), int
    _starting_points    = config.cvar("crazyranks_player_starting_points",   1000,      "The amount of points a player gets when he first connects"), int
    _do_bots_count      = config.cvar("crazyranks_do_bots_count",            1,         "Do bots count for ranks?"), bool
    _show_bots_on_top   = config.cvar("crazyranks_show_bots_on_topmenu",     0,         "Show bots on the top list? The setting above must be set to 1 for this to function"), bool
    _ranking_method     = config.cvar("crazyranks_ranking_method",           "steam",   "How should players be stored in the database (by steam, ip or name)"), str
    
    config.text(" ")
    config.text('*****************************')
    config.text('Command Settings')
    config.text('******************************')
      
    _rank_command       = config.cvar("crazyranks_command_rank",        "rank",         "Command that shows the player & others his positions and general information"), str
    _stats_command      = config.cvar("crazyranks_command_statsme",     "statsme",      "Command that shows the player detailed stats about himself"), str
    _next_command       = config.cvar("crazyranks_command_next",        "next",         "Command that shows the player the players that are infront of him and how many points they have"), str
    _hitbox_command     = config.cvar("crazyranks_command_hitbox",      "hitboxme",     "Command that shows the player detailed statistics about his hits"), str
    _weapon_command     = config.cvar("crazyranks_command_weapons",     "weaponme",     "Command that shows the player detailed statistics about his weapon usage and hits"), str
    
    config.text(" ")
    config.text('*****************************')
    config.text('Bonus Points Settings')
    config.text('*****************************')
    
    _headshot_bonus          = config.cvar("crazyranks_headshot_bonus",            5,      "The amount of extra points a player gets if he kills someone with a headshot"), int
    _bomb_plant_bonus        = config.cvar("crazyranks_bomb_plant_bonus",          10,     "The amount of extra points a player gets for planting the bomb"), int
    _bomb_explode_bonus      = config.cvar("crazyranks_bomb_explode_bonus",        15,     "The amount of extra points a player gets for the bomb exploding"), int
    _bomb_defuse_bonus       = config.cvar("crazyranks_bomb_defuse_bonus",         15,     "The amount of extra points a player gets for defusing the bomb"), int
    _hostage_rescue_bonus    = config.cvar("crazyranks_hostage_rescue_bonus",      15,     "The amount of extra points a player gets for rescueing a hostage"), int
    _round_end_winner_bonus  = config.cvar("crazyranks_end_round_winner_bonus",    15,     "The amount of extra points the winning team gets"), int
    
    config.text(" ")
    config.text('*****************************')
    config.text('Messaging Settings')
    config.text('*****************************')
    
    _message_player_events          = config.cvar("crazyranks_message_player_events",      1,      "Tell the player of his gain of points (e.g. on kill, bomb plant, hostage rescue...)"), bool
    _message_hurt_events            = config.cvar("crazyranks_message_hurt_events",        1,      "Show the player a detailed message of his hit on an enemy in the chatbox"), bool
    _message_player_connect_event   = config.cvar("crazyranks_message_on_player_connect",  1,      "Show a message in the chatbox when a player connects with his points"), bool
    
    config.text(" ")
    config.text('*****************************')
    config.text('Weapon Multipliers Settings')
    config.text('*****************************')
    
    _weapon_knife           = config.cvar("crazyranks_weapon_knife_multiplier",         4.0,    None), float
    _weapon_hegrenade       = config.cvar("crazyranks_weapon_hegrenade_multiplier",     2.0,    None), float
    _weapon_flashbang       = config.cvar("crazyranks_weapon_flashbang_multiplier",     5.0,    None), float
    _weapon_smokegrenade    = config.cvar("crazyranks_weapon_smokegrenade_multiplier",  5.0,    None), float
    _weapon_glock           = config.cvar("crazyranks_weapon_glock_multiplier",         1.3,    None), float
    _weapon_usp             = config.cvar("crazyranks_weapon_usp_multiplier",           1.3,    None), float
    _weapon_p228            = config.cvar("crazyranks_weapon_p228_multiplier",          1.4,    None), float
    _weapon_deagle          = config.cvar("crazyranks_weapon_deagle_multiplier",        1.3,    None), float
    _weapon_fiveseven       = config.cvar("crazyranks_weapon_fiveseven_multiplier",     1.4,    None), float
    _weapon_elite           = config.cvar("crazyranks_weapon_elite_multiplier",         1.4,    None), float
    _weapon_m3              = config.cvar("crazyranks_weapon_m3_multiplier",            1.1,    None), float
    _weapon_xm1014          = config.cvar("crazyranks_weapon_xm1014_multiplier",        1.1,    None), float
    _weapon_tmp             = config.cvar("crazyranks_weapon_tmp_multiplier",           1.3,    None), float
    _weapon_mac10           = config.cvar("crazyranks_weapon_mac10_multiplier",         1.3,    None), float
    _weapon_ump45           = config.cvar("crazyranks_weapon_ump45_multiplier",         1.2,    None), float
    _weapon_mp5navy         = config.cvar("crazyranks_weapon_mp5navy_multiplier",       1.1,    None), float
    _weapon_p90             = config.cvar("crazyranks_weapon_p90_multiplier",           1.1,    None), float
    _weapon_famas           = config.cvar("crazyranks_weapon_famas_multiplier",         1.1,    None), float
    _weapon_galil           = config.cvar("crazyranks_weapon_galil_multiplier",         1.1,    None), float
    _weapon_ak47            = config.cvar("crazyranks_weapon_ak47_multiplier",          1.0,    None), float
    _weapon_m4a1            = config.cvar("crazyranks_weapon_m4a1_multiplier",          1.0,    None), float
    _weapon_aug             = config.cvar("crazyranks_weapon_aug_multiplier",           1.0,    None), float
    _weapon_sg552           = config.cvar("crazyranks_weapon_sg552_multiplier",         1.0,    None), float
    _weapon_scout           = config.cvar("crazyranks_weapon_scout_multiplier",         1.1,    None), float
    _weapon_m249            = config.cvar("crazyranks_weapon_m249_multiplier",          1.0,    None), float
    _weapon_awp             = config.cvar("crazyranks_weapon_awp_multiplier",           1.3,    None), float
    _weapon_sg550           = config.cvar("crazyranks_weapon_sg550_multiplier",         1.2,    None), float
    _weapon_g3sg1           = config.cvar("crazyranks_weapon_g3sg1_multiplier",         1.2,    None), float
    
    config.execute()
    
    def __init__(self):
        self.config.write()
        
    def __getattr__(self, name):
        if hasattr(Settings, name):
            return super(Settings, self).__getattribute__(name)
           
        elif hasattr(Settings, '_'+name):
            val, func = super(Settings, self).__getattribute__('_'+name)
            return func(val)
   
    def __setattr__(self, name, value):
        if hasattr(Settings, name):
            super(Settings, self).__setattr__(name, value)
           
        elif hasattr(Settings, '_'+name):
            val, func = super(Settings, self).__getattribute__('_'+name)
            val.set(value)

class INIFile(object):
    def __init__(self, directory):
        self.path = directory
        if path(directory).isfile():
            self.exists = True
        else:
            self.exists = False
            open(directory, 'w').close()
            self.addInitialComments()
        
    def addInitialComments(self):
        header = [
        '# To add translations, use the language\'s abbrivation, for example:\n',
        '\n',
        '# en - English\n',
        '# de - German\n',
        '# fr - French\n',
        '# ru - Russian\n',
        '\n',
        '# There are more of them, just can\'t remember them\n',
        '\n',
        '# DO NOT CHANGE ANY WORD WITH $ IN IT!\n',
        '# You can have multiple colours. You can only use:\n',
        '\n',
        '# #default, #green and #lightgreen in one sentence!\n',
        '# #default, #team, #green in one sentence!\n',
        '\n',
        '# Special thanks to L\'In20Cible for the french translation. :)\n',
        '# Special thanks to TheCheeTaH for the danish translation. :)\n',
        '# Special thanks to Glazz for the portuguese translation. :)\n',
        '# Special thanks to chameleon82 for the russian translation. :)\n',]
        
        for line in header:
            stream = open(self.path, 'a')
            stream.write(line)
            
        stream.close()
        
    def addLanguageGroup(self, name, default, description, variables, abbr="en"):
        """
        Adds a language group to the translations file if it doesn't exists
        
        @param name: Name of group, str
        @param default: Default string in default language, str
        @param description: Short description of the language group, str
        @param variables: The variables and short description for explanation, dict
        @param abbr: Abbrevation of the language, str
        """
        
        if not self.exists:
            stream = open(self.path, 'a')
            
            stream.write('\n')
            stream.write('[' + name + ']\n')
            stream.write('# ' + description + '\n')
            if variables:
                stream.write('# \n')
                for variable in variables:
                    stream.write('# %s - %s\n' % (variable, variables[variable]))
            stream.write('\n')
            stream.write(abbr + '="' + default + '"\n')
            
            stream.close()
                
    def addMultipleLanguageGroups(self, groups=[], abbr="en"):
        """
        Adds language groups to the translations file if the file doesn't exists
        
        @param name: Name of group, str
        @param groups: A list of multiple language groups, list
        @param abbr: Abbrevation of the language, str
        """
        for group in groups:
            name, default, description, variables = group
            self.addLanguageGroup(name, default, description, variables)
                
    def getLanguageFile(self):
        return langlib.Strings(self.path)

settings = Settings()
translations = INIFile(addon_path + "/config/languages.ini")

# All weapons ...
all_weapons = ['ak47', 'm4a1', 'deagle', 'glock', 'usp',
   'p228', 'fiveseven', 'elite', 'm3', 'xm1014',
   'tmp', 'mac10', 'ump45', 'mp5navy', 'p90', 'famas', 'galil',
   'aug', 'sg552', 'scout', 'm249', 'awp',
   'sg550', 'g3sg1', 'knife', 'hegrenade', 'flashbang', 'smokegrenade', ]

weapon_convert = {
    'knife' : settings.weapon_knife,
    'hegrenade' : settings.weapon_hegrenade,
    'flashbang' : settings.weapon_flashbang,
    'smokegrenade' : settings.weapon_smokegrenade,
    'glock' : settings.weapon_glock,
    'usp' : settings.weapon_usp,
    'p228' : settings.weapon_p228,
    'deagle' : settings.weapon_deagle,
    'fiveseven' : settings.weapon_fiveseven,
    'elite' : settings.weapon_elite,
    'm3' : settings.weapon_m3,
    'xm1014' : settings.weapon_xm1014,
    'tmp' : settings.weapon_tmp,
    'mac10' : settings.weapon_mac10,
    'ump45' : settings.weapon_ump45,
    'mp5navy' : settings.weapon_mp5navy,
    'p90' : settings.weapon_p90,
    'famas' : settings.weapon_famas,
    'galil' : settings.weapon_galil,
    'ak47' : settings.weapon_ak47,
    'm4a1' : settings.weapon_m4a1,
    'aug' : settings.weapon_aug,
    'sg552' : settings.weapon_sg552,
    'scout' : settings.weapon_scout,
    'm249' : settings.weapon_m249,
    'awp' : settings.weapon_awp,
    'sg550' : settings.weapon_sg550,
    'g3sg1' : settings.weapon_g3sg1, }

translations.addMultipleLanguageGroups([
    ['steamid command',
     "#greenRanks: #defaultYou are being ranked based on your #green$method#default, which is: #team$value",
     'Displays a message showing the player on which method he is being saved in the database.',
     {'$method' : 'The method he is being saved in the database (IP, STEAMID, NAME)', '$value' : 'Either the ip, steamid or name of the player'},]
    ,
    ['player connect',
     "#greenRanks: #defaultPlayer #team$player #defaultconnected with #green$points#default (Pos #team$position#default of #team$total#default) points.",
     'Displays a message when the player connects with his points and position.',
     {'$player' : 'The connecting players name', '$points' : 'His current points count', '$position' : 'The players position in the ranks', '$total' : 'Total players ranked by the database'}]
    ,
    ['round end',
     "#greenRanks: #team$team#default have received #green$bonus#default points for winning the round.",
     'Displays a message at round end to all players showing who won the round.',
     {'$team' : 'The team that won the round', '$bonus' : 'The amount of points they got for winning'}]
    ,
    ['player death no headshot',
     "#greenRanks: #default$attacker (#green$points#default) gained #green$bonus#default points for killing #team$victim #default($victimpoints)",
     'Shows a message to the attacker when he kills someone.',
     {'$attacker' : 'The player that killed', '$points' : 'The amount of points the attacker has', '$bonus' : 'The amount of points he got for killing ther player', '$victim' : 'The player that was killed', '$victimpoints' : 'The amount of points the victim has'}]
    ,
    ['player death with headshot',
     "#greenRanks: #default$attacker (#green$points#default) gained #green$bonus#default [+$hs_bonus hs] points for killing #team$victim #default($victimpoints)",
     'Shows a message to the attacker when he kills someone with a headshot.',
     {'$attacker' : 'The player that killed', '$points' : 'The amount of points the attacker has', '$bonus' : 'The amount of points he got for killing ther player', '$hs_bonus' : 'The extra points he gets for headshot', '$victim' : 'The player that was killed', '$victimpoints' : 'The amount of points the victim has'}]
    ,
    ['player death victim notify no headshot',
     "#greenRanks: #team$attacker#default (#green$points#default) gained #green$bonus#default points for killing $victim ($victimpoints)",
     'Shows the victim a message when he gets killed.',
     {'$attacker' : 'The player that killed', '$points' : 'The amount of points the attacker has', '$bonus' : 'The amount of points he got for killing ther player', '$victim' : 'The player that was killed', '$victimpoints' : 'The amount of points the victim has'}]
    ,
    ['player death victim notify with headshot',
     "#greenRanks: #team$attacker#default (#green$points#default) gained #green$bonus#default [+$hs_bonus hs] points for killing $victim ($victimpoints)",
     'Shows the victim a message when he gets killed with a headshot.',
     {'$attacker' : 'The player that killed', '$points' : 'The amount of points the attacker has', '$bonus' : 'The amount of points he got for killing ther player', '$hs_bonus' : 'The extra points he gets for headshot', '$victim' : 'The player that was killed', '$victimpoints' : 'The amount of points the victim has'}]
    ,
    ['hostage rescue',
     "#greenRanks: #defaultYou gained #green$bonus#default (#green$points#default) points for #teamrescuing a hostage#default.",
     'Shows a message to the player when he saves a hostage.',
     {'$bonus' : 'The extra amount of points he gets.', '$points' : 'The amount of points the player has'}]
    ,
    ['bomb planted',
     "#greenRanks: #defaultYou gained #green$bonus#default (#green$points#default) points for #teamplanting the bomb#default.",
     'Shows a message to the player when he plants the bomb.',
     {'$bonus' : 'The extra amount of points for planting the bomb.', '$points' : 'The amount of points the player has'}]
    ,
    ['bomb defused',
     "#greenRanks: #defaultYou gained #green$bonus#default (#green$points#default) points for #teamdefusing the bomb#default.",
     'Shows a message to the player when he defuses a bomb.',
     {'$bonus' : 'The extra amount of points for defusing the bomb.', '$points' : 'The amount of points the player has'}]
    ,
    ['bomb exploded',
     "#greenRanks: #defaultYou gained #green$bonus#default (#green$points#default) points for #teamthe bomb exploding#default.",
     'Shows a message to the player when the bomb explodes.',
     {'$bonus' : 'The extra amount of points for the bomb exploding.', '$points' : 'The amount of points the player has'}]
    ,
    ['player hurt',
     "#greenRanks: #defaultYou hurt player #team$victim#default in his #team$hitbox#default (damage: #green$damage#default)",
     'Shows a message to the attacker when he hits a player.',
     {'$victim' : 'The player that was hit', '$hitbox' : 'The hitbox where the hit was impacted', '$damage' : 'The amount of damage done'}]
    ,
    ['player without points',
     "#defaultRank #green$position#default/#green$total#default: #team$player #defaultwith #green$kills#default kills, #green$deaths#default deaths, kdr #green$kdr",
     'Shows a short statistic message to the player when he triggers the command.',
     {'$position' : 'The position the player is at', '$total' : 'The total amount of players ranked', '$player' : 'The name of the player', '$kills' : 'The amount of kills he has', '$deaths' : 'The amount of times he died', '$kdr' : 'The k/d ratio of the player'}]
    ,
    ['player with points',
     "#defaultRank #green$position#default/#green$total#default: #team$player #defaultwith #green$points#default points, #green$kills#default kills, #green$deaths#default deaths, kdr #green$kdr",
     'Shows a short statistic message to the player when he triggers the command.',
     {'$position' : 'The position the player is at', '$total' : 'The total amount of players ranked', '$player' : 'The name of the player', '$points' : 'The amount of points the player has', '$kills' : 'The amount of kills he has', '$deaths' : 'The amount of times he died', '$kdr' : 'The k/d ratio of the player'}]
    ,
    ])

text = translations.getLanguageFile()
# Tell line
def tell(userid, index, lang_text, lang_args={}):
    if int(userid):
        saytext2(userid, index, text(lang_text, lang_args, playerlib.getPlayer(userid).get('lang')))
    
def msg(index, lang_text, lang_args={}):
    for userid in es.getUseridList():
        saytext2(userid, index, text(lang_text, lang_args, playerlib.getPlayer(userid).get('lang')))

''' Class Modules '''

########################################
##
## DATABASE CLASS

class DataBase:
    ''' The ranking database, handles all off the data '''
    def __init__(self):
        ''' Fires when the script is initialized '''
        
        # Get the points_system value, so it can be accessed easily
        self.points = settings.points_system
        # Create a connection to our database
        self.connection = sqlite3.connect(addon_path + '/data/ranks.sqlite3')
        # Establish a cursor for data handling
        self.cursor = self.connection.cursor()
        # Set the journal mode to off
        self.execute("PRAGMA journal_mode=OFF")
        # Set the locking mode to exclusive
        self.execute("PRAGMA locking_mode=EXCLUSIVE")
        # Create the main table with our player kills/deaths/suicides/damage...
        self.execute("CREATE TABLE IF NOT EXISTS data (steamid TEXT PRIMARY KEY, name TEXT, kills INTEGER UNSIGNED, deaths INTEGER UNSIGNED, damage INTEGER UNSIGNED, tWins INTEGER UNSIGNED, ctWins INTEGER UNSIGNED, ctTotal INTEGER UNSIGNED, tTotal INTEGER UNSIGNED, bP INTEGER UNSIGNED, bD INTEGER UNSIGNED, hR INTEGER UNSIGNED, hK INTEGER UNSIGNED, bE INTEGER UNSIGNED, suicides INTEGER UNSIGNED, tP INTEGER UNSIGNED, hs INTEGER UNSIGNED, points INTEGER UNSIGNED)")
        # Create our secondary table with our players hitbox hits
        self.execute("CREATE TABLE IF NOT EXISTS hitbox (steamid TEXT PRIMARY KEY, head INTEGER UNSIGNED, chest INTEGER UNSIGNED, leftarm INTEGER UNSIGNED, rightarm INTEGER UNSIGNED, stomach INTEGER UNSIGNED, leftleg INTEGER UNSIGNED, rightleg INTEGER UNSIGNED)")

        string = "CREATE TABLE IF NOT EXISTS weapons (steamid TEXT PRIMARY KEY"
        for weapon in all_weapons:
            string += ', %s INTEGER UNSIGNED' % weapon
            if weapon.startswith('smoke'):
                string += ')'
        self.execute(string)
        
        # Create our third table with the players names
        self.execute("CREATE TABLE IF NOT EXISTS names (steamid TEXT PRIMARY KEY, name TEXT)")
        
    ##
    ## MAIN DATABASE FUNCTIONS
    ##
  
    def execute(self, sql, args=None):
        ''' Executes certain actions in the database '''
        if args:
            self.cursor.execute(sql, args)
        else:
            self.cursor.execute(sql)
       
    def fetchone(self, sql):
        ''' Fetches data from the table from a single row '''
        self.cursor.execute(sql)
        return self.cursor.fetchone()
       
    def fetchall(self, sql):
        ''' Fetches data from the table from all rows '''
        self.cursor.execute(sql)
        return self.cursor.fetchall()
        
    def __delitem__(self, entry):
        self.execute('DELETE FROM data WHERE steamid="%s"' % entry)
        

    ##
    ## USER EXISTENCE CHECK FUNCTIONS
    ##
       
    def __contains__(self, steamid):
        ''' If the script contains the players steamid then return True otherwise False '''
        return bool(self.fetchone('SELECT steamid FROM data WHERE steamid="%s"'%steamid))
       
    def hitCheck(self, steamid):
        ''' Checks whether the player is in the hitbox database '''
        return bool(self.fetchone('SELECT steamid FROM hitbox WHERE steamid="%s"'%steamid))
        
    def weaponCheck(self, steamid):
        ''' Checks whether the player is in the weapons database '''
        return bool(self.fetchone('SELECT steamid FROM weapons WHERE steamid="%s"'%steamid))
        
    def nameCheck(self, steamid):
        ''' Checks whether the player is in the names database '''
        return bool(self.fetchone('SELECT steamid FROM names WHERE steamid="%s"'%steamid))
        
        
    ##
    ## GET FUNCTIONS
    ##
    
    def createNameDatabase(self):
        ''' Creates our name database if the update was after the first creation of the database '''
        self.execute("CREATE TABLE IF NOT EXISTS names (steamid TEXT PRIMARY KEY, name TEXT)")
        
    def getPlayerName(self, steamid):
        ''' Gets the players name from the database '''
        if not steamid: return
        if not self.nameCheck(steamid): self.setPlayerName(steamid, es.getplayername(es.getuserid(steamid)))
        result = self.fetchone("SELECT name FROM names WHERE steamid='%s'"%steamid)
        if not result:
            return None
          
        return result[0]
        
    def getBotName(self, steamid):
        ''' Gets the bots name from the database '''
        if steamid in self:
            result = self.fetchone("SELECT name FROM data WHERE steamid='%s'"%steamid)
            if not result:
                return None
              
            return result[0]
       
    def getDamage(self, steamid):
        ''' Gets the players damage '''
        result = self.fetchone("SELECT damage FROM data WHERE steamid='%s'"%steamid)
        if not result:
            return None
          
        return result
 
    def getHitboxData(self, steamid):
        ''' Gets all of the players hitbox data (head hits, chest hits...) '''
        result = self.fetchone("SELECT head, chest, stomach, leftarm, rightarm, leftleg, rightleg FROM hitbox WHERE steamid='%s'"%steamid)
        if not result:
            return None
          
        return result
    
    def getTop10WeaponData(self, steamid):
        ''' Returns the weapons for the user '''
        result = self.fetchone("SELECT * FROM weapons WHERE steamid='%s' DESC LIMIT 10"%steamid)
        if not result:
            return None
          
        return result
    
    def getWeaponData(self, steamid):
        ''' Returns the weapons for the user '''
        result = self.fetchone("SELECT * FROM weapons WHERE steamid='%s'"%steamid)
        if not result:
            return None
          
        return result
       
    def getTotalCount(self):
        ''' Get the total count of players in the database '''
        result = self.fetchone("SELECT count(*) FROM data")
        if result:
            return result[0]
              
        return None
    
    def getPoints(self, steamid):
        ''' Gets the players total points '''
        result = self.fetchone("SELECT points FROM data WHERE steamid='%s'"%steamid)
        if not result:
            return None
          
        return result[0]
    
    def getKills(self, steamid):
        ''' Gets the players total kills '''
        result = self.fetchone("SELECT kills FROM data WHERE steamid='%s'"%steamid)
        if not result:
            return None
          
        return result[0]
    
    def getAll(self):
        ''' Get all of the players in the database '''
        if not self.points:
            result = self.fetchall("SELECT steamid, kills, deaths FROM data ORDER BY kills DESC LIMIT %s" % self.getTotalCount())
        else:
            result = self.fetchall("SELECT steamid, kills, deaths, points FROM data ORDER BY points DESC LIMIT %s" % self.getTotalCount())
        
        if not result:
            return None
        
        return result
    
    def getPosition(self, steamid):
        ''' Get the players position (rank) '''
        self.cursor.execute("SELECT count(*) FROM data WHERE points<(SELECT points FROM data WHERE steamid=?)", (steamid,))
        result = self.cursor.fetchone()
        if result:
            return self.getTotalCount() - result[0]
            
    def getPlayerTime(self, steamid):
        ''' Gets the players game time '''
        result = self.fetchone("SELECT tP FROM data WHERE steamid='%s'"%steamid)
        if not result:
            return None
          
        return result[0]
            
    ##
    ## ADD FUNCTIONS
    ##
    
    def addWeaponKill(self, steamid, weapon):
        ''' Add the weapon kill to the players weapon statistics '''
        if not steamid: return
        if self.weaponCheck(steamid):
            # Add points to his current value
            self.execute('UPDATE weapons SET %s=%s+1 WHERE steamid="%s"' % (weapon, weapon, steamid))
        
        # If he isnt in the database..
        else:
            # And set his variables
            self.execute('INSERT INTO weapons VALUES (?, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)', (steamid,))

    
    def addTime(self, steamid, time):
        if not steamid: return
        if steamid in self:
            # Add points to his current value
            self.execute('UPDATE data SET tP=tP+%s WHERE steamid="%s"' % (time, steamid))
        
        # If he isnt in the database..
        else:
            # Get his name
            if steamid.startswith('BOT_'):
                name = steamid.replace('BOT_', '')
            else:
                name = es.getplayername(es.getuserid(steamid))
                  
            # And set his variables
            self.execute('INSERT INTO data VALUES (?, ?, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, ?, 0, 0)', (steamid, name, time))
     
    def addPoints(self, steamid, points):
        ''' Adds points to the player '''
        if not steamid: return
        # If the point system is enabled
        if self.points:
            # If he is in the database
            if steamid in self:
                # Add points to his current value
                self.execute('UPDATE data SET points=points+%s WHERE steamid="%s"' % (points, steamid))
    
            # If he isnt in the database..
            else:
                # Get his name
                if steamid.startswith('BOT_'):
                    name = steamid.replace('BOT_', '')
                else:
                    name = es.getplayername(self.getPlayerName(steamid))
                   
                # And set his variables
                self.execute('INSERT INTO data VALUES (?, ?, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 , 0, 0, 0, 0, 0, ?)', (steamid, name, int(settings.starting_points)))
                
    def addDefaultPoints(self, steamid):
        if not steamid: return
        if self.points:
            # If he is in the database
            if steamid not in self:
                # Get his name
                if steamid.startswith('BOT_'):
                    name = steamid.replace('BOT_', '')
                else:
                    name = es.getplayername(self.getPlayerName(steamid))
                   
                # And set his variables
                self.execute('INSERT INTO data VALUES (?, ?, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 , 0, 0, 0, 0, 0, ?)', (steamid, name, int(settings.starting_points)))
            
    ##
    ## SET FUNCTIONS
    ##
    
    def setPlayerName(self, steamid, name):
        if not steamid: return
        if self.nameCheck(steamid):
            # Add points to his current value
            self.execute('UPDATE names SET name="%s" WHERE steamid="%s"' % (name, steamid))
        
        # If he isnt in the database..
        else:
            # Dumb method, might change in future
            if steamid == 'BOT':
                steamid == name
            # Set his variables
            self.execute('INSERT INTO names VALUES (?, ?)', (steamid, name))
                  
    
    ##
    ## GAMEPLAY EVENT HANDLING FUNCTIONS
    ##
    
    def roundend(self, winner):
        ''' Updated the database at the end of the round '''
          
        # For each userid..
        for userid in es.getUseridList():
            # Get the team of the user
            team = es.getplayerteam(userid)
            # If the player is a bot
            steamid = getPlayer(userid).getUniqueId()
            if not steamid: continue
            # If he is not in the database
            if not steamid in self:
                # Get his name
                name = es.getplayername(userid)
                # Apply the data into the database
                self.execute('INSERT INTO data VALUES (?, ?, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, ?)', (steamid, name, int(settings.starting_points)))
                     
            # If he is a Terrorist
            if team == 2:
                # Set his total round + 1
                self.execute('UPDATE data SET tTotal=tTotal+1 WHERE steamid="%s"' % steamid)
                # If Terrorists were the winners...
                if winner == 2:
                    # Add a point to his total wins
                    self.execute('UPDATE data SET tWins=tWins+1 WHERE steamid="%s"' % steamid)
                    if self.points:
                        # Add points
                        self.addPoints(steamid, settings.round_end_winner_bonus)
                    
            # If the player is a CT
            elif team == 3:
                # Set his total rounds + 1
                self.execute('UPDATE data SET ctTotal=ctTotal+1 WHERE steamid="%s"' % steamid)
                # If CTs were the winners
                if winner == 3:
                    # Update the database
                    self.execute('UPDATE data SET ctWins=ctWins+1 WHERE steamid="%s"' % steamid)
                    # Add points...
                    self.addPoints(steamid, settings.round_end_winner_bonus)
                      
    def hurt(self, attacker, damage, userid):
       ''' Updates the database when a player is hurt '''
       if attacker:
       
            # If the attacker is in the database
            if attacker in self:
                # Update the database with damage
                self.execute('UPDATE data SET damage=damage+%s WHERE steamid="%s"' % (damage, attacker))
            # If he isn't...
            else:
            # Get his name
                if attacker.startswith('BOT_'):
                    name = attacker.replace('BOT_', '')
                else:
                    name = self.getPlayerName(getPlayer(userid).getUniqueId())
                  
                # Set his table details
                self.execute('INSERT INTO data VALUES (?, ?, 0, 0, ?, 0, 0, 0, 0, 0, 0, 0 , 0, 0, 0, 0, 0, ?)', (attacker, name, damage, int(settings.starting_points)))
           
       
    def hitbox(self, attacker, hit):
        ''' Updated the 'hitbox' database when a player is hurt '''
        if not attacker: return
        # If the player isn't in the database
        if not self.hitCheck(attacker):
            # Create him...
            self.execute('INSERT INTO hitbox VALUES (?, 0, 0, 0, 0, 0, 0, 0)', (attacker,))
        
        # If it's a headshot
        if hit == 1:
            # Update the database...
            self.execute('UPDATE hitbox SET head=head+1 WHERE steamid="%s"' % attacker)
            
        # If it's a chest hit
        elif hit == 2:
            # Update the database...
            self.execute('UPDATE hitbox SET chest=chest+1 WHERE steamid="%s"' % attacker)
            
        # If it's a stomach hit
        elif hit == 3:
            # Update the database...
            self.execute('UPDATE hitbox SET stomach=stomach+1 WHERE steamid="%s"' % attacker)
            
        # If it's a left arm hit
        elif hit == 4:
            # Update the database...
            self.execute('UPDATE hitbox SET leftarm=leftarm+1 WHERE steamid="%s"' % attacker)
            
        # If it's a right arm hit
        elif hit == 5:
            # Update the database...
            self.execute('UPDATE hitbox SET rightarm=rightarm+1 WHERE steamid="%s"' % attacker)
            
        # If it's a left leg hit
        elif hit == 6:
            # Update the database...
            self.execute('UPDATE hitbox SET leftleg=leftleg+1 WHERE steamid="%s"' % attacker)
            
        # If it's a right leg hit
        elif hit == 7:
            # Update the database...
            self.execute('UPDATE hitbox SET rightleg=rightleg+1 WHERE steamid="%s"' % attacker)
               
   
    def death(self, victim, attacker, weapon, headshot=False, victim_userid=None, attacker_userid=None):
        ''' Updated the database when a player dies '''
        if not attacker: return
        if not victim: return
        # If the attacker is equal to the victim
        if attacker == victim:
            # If he exists...
            if attacker in self:
                # He did a sucide
                self.execute('UPDATE data SET suicides=suicides+1 WHERE steamid="%s"' % attacker)
               
           # If he isn't in the database
            else:
                # Get his name
                if attacker.startswith('BOT_'):
                    name = attacker.replace('BOT_', '')
                else:
                    name = self.getPlayerName(getPlayer(attacker_userid).getUniqueId())
                # Create him
                self.execute('INSERT INTO data VALUES (?, ?, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 , 0, 0, 1, 0, 0, ?)', (attacker, name, int(settings.starting_points)))
        
        # If the attacker is not equal to the victim and the attacker exists (world check)
        if attacker != victim and attacker:
            # If he exists...
            if attacker in self:
                # Update his kills
                self.execute('UPDATE data SET kills=kills+1 WHERE steamid="%s"' % attacker)
            
            # If he doesn't exist
            else:
                # Get his name
                if attacker.startswith('BOT_'):
                    name = attacker.replace('BOT_', '')
                else:
                    name = self.getPlayerName(getPlayer(attacker_userid).getUniqueId())
                # Update the database
                self.execute('INSERT INTO data VALUES (?, ?, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 , 0, 0, 0, 0, 0, ?)', (attacker, name, int(settings.starting_points)))
                
            # If it was a headshot
            if headshot:
                # Add a point to his headshot kills
                self.execute('UPDATE data SET hs=hs+1 WHERE steamid="%s"' % attacker)
                
        # If the victim exists
        if victim in self:
            # Add 1 to his deaths
            self.execute('UPDATE data SET deaths=deaths+1 WHERE steamid="%s"' % victim)
           
        # If he doesnt..
        else:
            # Get his name
            if victim.startswith('BOT_'):
                name = victim.replace('BOT_', '')
            else:
                name = self.getPlayerName(getPlayer(victim_userid).getUniqueId())
            # Create him
            self.execute('INSERT INTO data VALUES (?, ?, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, ?)', (victim, name, int(settings.starting_points)))
            
        # If the points system is running
        if self.points:
            if attacker != victim and attacker:
                # Get the victim points
                victim_points = self.getPoints(victim)
                # Get the attacker points
                attacker_points = self.getPoints(attacker)
                # Do the formula
                bonus = int((victim_points/attacker_points)*settings.multiplier*float(weapon_convert[weapon]))
                     
                # If it was a headshot
                if headshot:
                    # Add the headshot bonus
                    bonus = bonus + settings.headshot_bonus
               
                # Add the number of points he earned
                self.addPoints(attacker, bonus)
               
                if victim_userid and attacker_userid:
                    index = playerlib.getPlayer(victim_userid).attributes['index']
                    point = self.getPoints(attacker)
                    victim_point = self.getPoints(victim)
                    if not headshot:
                        if attacker_userid: tell(attacker_userid, index, 'player death no headshot', {'attacker' : self.getPlayerName(getPlayer(attacker_userid).getUniqueId()), 'points' : getFormat(point), 'bonus' : str(bonus), 'victim' : self.getPlayerName(getPlayer(victim_userid).getUniqueId()), 'victimpoints' : getFormat(victim_point)})
                    else:
                        if attacker_userid: tell(attacker_userid, index, 'player death with headshot', {'attacker' : self.getPlayerName(getPlayer(attacker_userid).getUniqueId()), 'points' : getFormat(point), 'bonus' : str(bonus), 'hs_bonus' : settings.headshot_bonus, 'victim' : self.getPlayerName(getPlayer(victim_userid).getUniqueId()), 'victimpoints' : getFormat(victim_point)})
                                
                index = playerlib.getPlayer(attacker_userid).attributes['index']
                point = self.getPoints(attacker)
                victim_point = self.getPoints(victim)
                if not headshot:
                    if victim_userid: tell(victim_userid, index, 'player death victim notify no headshot', {'attacker' : self.getPlayerName(getPlayer(attacker_userid).getUniqueId()), 'points' : getFormat(point), 'bonus' : str(bonus), 'victim' : self.getPlayerName(getPlayer(victim_userid).getUniqueId()), 'victimpoints' : getFormat(victim_point)})
                else:
                    if victim_userid: tell(victim_userid, index, 'player death victim notify with headshot', {'attacker' : self.getPlayerName(getPlayer(attacker_userid).getUniqueId()), 'points' : getFormat(point), 'bonus' : str(bonus), 'hs_bonus' : settings.headshot_bonus, 'victim' : self.getPlayerName(getPlayer(victim_userid).getUniqueId()), 'victimpoints' : getFormat(victim_point)})
            
        if weapon in all_weapons:
            self.addWeaponKill(attacker, weapon)
    
    def hostage(self, steamid, reason, userid):
        ''' Updated the database when a hostage event occurs '''
        # 0 = Killed, 1 = Rescued
        
        # If the hostage was killed,
        # Add the player to the database if he doesnt already exist
        # and update his hostage kills.
        if steamid:
            if reason == 0:
                if steamid in self:
                    self.execute('UPDATE data SET hK=hK+1 WHERE steamid="%s"' % steamid)
                else:
                    # Get his name
                    if steamid.startswith('BOT_'):
                        name = steamid.replace('BOT_', '')
                    else:
                        name = self.getPlayerName(getPlayer(userid).getUniqueId())
                    
                    self.execute('INSERT INTO data VALUES (?, ?, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, ?)', (steamid, name, int(settings.starting_points)))
                   
           # If the hostage was rescued,
           # Add the player to the database if he doesnt already exist
           # and update his hostage rescued.    
            elif reason == 1:
                if steamid in self:
                    self.execute('UPDATE data SET hR=hR+1 WHERE steamid="%s"' % steamid)
                else:
                    # Get his name
                    if steamid.startswith('BOT_'):
                        name = steamid.replace('BOT_', '')
                    else:
                        name = self.getPlayerName(getPlayer(userid).getUniqueId())
                    
                    self.execute('INSERT INTO data VALUES (?, ?, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, ?)', (steamid, name, int(settings.starting_points)))
                    
                # Add the bonus hostage rescue points
                self.addPoints(steamid, settings.hostage_rescue_bonus)
               
                # If we should notify the player of his gain of points
                if settings.message_player_events:
                    if not steamid.startswith('BOT_'):
                        index = playerlib.getPlayer(userid).attributes['index']
                        points = self.getPoints(steamid)
                        tell(userid, index, 'hostage rescue', {'bonus' : settings.hostage_rescue_bonus, 'points' : getFormat(points)})
               
               
    def bomb(self, steamid, reason, userid):
        ''' Updated the database when a bomb event occurs '''
        # 0 = Planted, 1 = Defused, 2 = Exploded
        if not steamid: return
        # If the bomb was planted,
        # Add the player to the database if he doesnt already exist
        # and update his bomb plants and add the bonus points.
        if reason == 0:
            if steamid in self:
                self.execute('UPDATE data SET bP=bP+1 WHERE steamid="%s"' % steamid)
            else:
                # Get his name
                if steamid.startswith('BOT_'):
                    name = steamid.replace('BOT_', '')
                else:
                    name = self.getPlayerName(getPlayer(userid).getUniqueId())
                
                self.execute('INSERT INTO data VALUES (?, ?, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, ?)', (steamid, name, int(settings.starting_points)))
                
            self.addPoints(steamid, settings.bomb_plant_bonus)
           
            # If we should notify the player of his gain of points
            if settings.message_player_events:
                if not steamid.startswith('BOT_'):
                    index = playerlib.getPlayer(userid).attributes['index']
                    points = self.getPoints(steamid)
                    tell(userid, index, 'bomb planted', {'bonus' : settings.bomb_plant_bonus, 'points' : getFormat(points)}) 
                
        # If the bomb was defused,
        # Add the player to the database if he doesnt already exist
        # and update his bomb defuses and add the bonus points.
        elif reason == 1:
            if steamid in self:
                self.execute('UPDATE data SET bD=bD+1 WHERE steamid="%s"' % steamid)
            else:
                # Get his name
                if steamid.startswith('BOT_'):
                    name = steamid.replace('BOT_', '')
                else:
                    name = self.getPlayerName(getPlayer(userid).getUniqueId())
               
                self.execute('INSERT INTO data VALUES (?, ?, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, ?)', (steamid, name, int(settings.starting_points)))
                
            self.addPoints(steamid, settings.bomb_defuse_bonus)
           
            # If we should notify the player of his gain of points
            if settings.message_player_events:
                if not steamid.startswith('BOT_'):
                    index = playerlib.getPlayer(userid).attributes['index']
                    points = self.getPoints(steamid)
                    tell(userid, index, 'bomb defused', {'bonus' : settings.bomb_defuse_bonus, 'points' : getFormat(points)}) 
                
        # If the bomb exploded,
        # Add the player to the database if he doesnt already exist
        # and update his bomb explode count and add the bonus points.
        elif reason == 2:
            if steamid in self:
                self.execute('UPDATE data SET bE=bE+1 WHERE steamid="%s"' % steamid)
            else:
                # Get his name
                if steamid.startswith('BOT_'):
                    name = steamid.replace('BOT_', '')
                else:
                    name = self.getPlayerName(getPlayer(userid).getUniqueId())
               
                self.execute('INSERT INTO data VALUES (?, ?, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, ?)', (steamid, name, int(settings.starting_points)))
                
            self.addPoints(steamid, settings.bomb_explode_bonus)
            
            # If we should notify the player of his gain of points
            if settings.message_player_events:
                if not steamid.startswith('BOT_'):
                    index = playerlib.getPlayer(userid).attributes['index']
                    points = self.getPoints(steamid)
                    tell(userid, index, 'bomb exploded', {'bonus' : settings.bomb_explode_bonus, 'points' : getFormat(points)}) 

    ##
    ## STATISTIC FUNCTIONS
    ##
    
    def top10(self):
        ''' Gets the top 10 players '''
        if not self.points:
            result = self.fetchall("SELECT steamid, kills, deaths FROM data ORDER BY kills DESC LIMIT 10")
        else:
            result = self.fetchall("SELECT steamid, kills, deaths, points FROM data ORDER BY points DESC LIMIT 10")
        if not result:
            return None
        
        return result
        # ((steamid, kills, deaths), (steamid, kills, deaths),...)
       
    def stats(self, steamid):
        ''' Gets most of the players info which is shown on the 'statsme' command '''
        result = self.fetchone("SELECT tWins, ctWins, ctTotal, tTotal, bP, bD, hR, hK, bE, suicides, hs, tP FROM data WHERE steamid='%s'"%steamid)
        if not result:
            return None
        
        return result
       
    def rank(self, steamid):
        ''' Get the players info which is required when he types 'rank' in the chat '''
        if not self.points:
            result = self.fetchone("SELECT kills, deaths FROM data WHERE steamid='%s'"%steamid)
        else:
            result = self.fetchone("SELECT kills, deaths, points FROM data WHERE steamid='%s'"%steamid)
        
        if not result:
            return None
        
        return result
    
    ##
    ## DATABASE SAVE & DISCONNECT FUNCTIONS
    ##
               
    def commit(self):
        ''' Save (commit) the changes '''
        self.connection.commit()
       
    def close(self, save=True):
        ''' Close the database '''
        if save:
            self.connection.commit()
        
        self.connection.close()
       
# Bind our database to a variable
database = DataBase()

players_conn = {}

################################################################
##
## Player Class

class Player(object):
    
    def __init__(self, playerid):
        self.userid = playerid
        self.steam = self.getUniqueId()
        
    def __getattr__(self, name):
        if hasattr(Player, name):
            return super(Player, self).__getattribute__(name)
           
        else:
            try:
                return self.get(name)
            except:
                es.dbgmsg(0, 'RANKS: There is no %s attribute in the Player class' % name)
        
    def get(self, name):
        if name == 'points':
            return database.getPoints(self.steam)
        elif name == 'rank':
            return database.rank(self.steam)
        elif name == 'index':
            return playerlib.getPlayer(self.userid).attributes['index']
        elif name == 'name':
            if self.steam.startswith('BOT_'):
                name = self.steam.replace('BOT_', '')
            else:
                name = database.getPlayerName(self.steam)
            return name
        elif name == 'timeplayed':
            return GetInHMS(database.getPlayerTime(self.steam))
        elif name == 'damage':
            return database.getDamage(self.steam)
        elif name == 'hitbox':
            return database.getHitboxData(self.steam)
        elif name == 'weapons':
            return database.getWeaponData(self.steam)
        elif name == 'kills':
            return database.getKills(self.steam)
        elif name == 'position':
            return database.getPosition(self.steam)
        elif name == 'statsme':
            return database.stats(self.steam)
        elif name == 'uniqueid':
            return self.getUniqueId()
            
    def set(self, name, value):
        if name == 'name':
            database.setPlayerName(self.steam, value)
        
    def add(self, name, value=None):
        if name == 'weaponkill':
            database.addWeaponKill(self.steam)
        elif name == 'timeplayed':
            database.addTime(self.steam, value)
        elif name == 'defaultpoints':
            database.addDefaultPoints(self.steam)
        elif name == 'points':
            database.addPoints(self.steam, value)
        
    def exists(self):
        if es.exists('userid', self.userid):
            return True
        
        return False
    
    def getUniqueId(self):
        # Checks whether the player is a bot and returns a unique id if he is,
        # otherwise returns a stemid
        if self.exists():
            player = playerlib.getPlayer(self.userid)
            if es.isbot(self.userid):
                if settings.do_bots_count:
                    return 'BOT_' + es.getplayername(self.userid)
            else:
                methods = {"steam" : player.steamid, "ip" : player.address[:player.address.find(':')], "name" : player.name}
                return methods[str(settings.ranking_method)]

    def showRank(self):
        if self.exists():
            index = playerlib.getPlayer(self.userid).attributes['index']
            if database.rank(self.steam) != None:
                if not settings.points_system:
                    kills, deaths = database.rank(self.steam)
                else:
                    kills, deaths, points = database.rank(self.steam)
                kdr = convertKdr(kills, deaths)
            else:
                es.tell(self.userid, '#green', 'You are not ranked yet.')
                return
        
            position = database.getPosition(self.steam)
            countAll = database.getTotalCount()
            if not settings.points_system:
                msg(index, 'rank without points', {'position' : position, 'total' : countAll, 'player' : es.getplayername(self.userid), 'kills' : kills, 'deaths' : deaths, 'kdr' : kdr})
            else:
                msg(index, 'rank with points', {'position' : position, 'total' : countAll, 'player' : es.getplayername(self.userid), 'points' : getFormat(points), 'kills' : kills, 'deaths' : deaths, 'kdr' : kdr})
                
    def showStats(self):
        if self.exists():
            if self.rank != None:
                if not settings.points_system:
                    kills, deaths = database.rank(self.steam)
                else:
                    kills, deaths, points = database.rank(self.steam)
                kdr = convertKdr(kills, deaths)
            else:
                es.tell(self.userid, '#green', 'You are not ranked yet.')
                return
            
            kdr = convertKdr(kills, deaths)
            tWins, ctWins, ctTotal, tTotal, bP, bD, hR, hK, bE, suicides, headshots, tP = self.statsme
            t_win_ratio, ct_win_ratio = getWinRatio(tWins, tTotal, ctWins, ctTotal)
            position = database.getPosition(self.steam)
            countAll = database.getTotalCount()
         
            stats = popuplib.create('stats_%s' % self.userid)
            stats.addline('%s - Total' % es.getplayername(self.userid))
            stats.addline(' ')
            stats.addline('->1. ) Player Ranking')
            stats.addline(' Position %s of %s' % (position, countAll))
            if settings.points_system:
                stats.addline(' %s Points' % getFormat(points))
            stats.addline(' %s : %s (%s)' % (kills, deaths, kdr))
            stats.addline(' %s Headshots (%s%s)' % (headshots, convertKdr(headshots, kills, True), '%'))
            stats.addline(' %s Suicides' % suicides)
            stats.addline(' %s' % GetInHMS(int(tP)))
            stats.addline(' ')
            stats.addline('->2. ) Bomb Statistics')
            stats.addline(' %s Planted,' % bP)
            stats.addline(' %s Defused,' % bD)
            stats.addline(' %s Exploded' % bE)
            stats.addline(' ')
            stats.addline('->3. ) Hostage Statistics')
            stats.addline(' %s Killed,' % hK)
            stats.addline(' %s Rescued' % hR)
            stats.addline(' ')
            stats.addline('->4. ) Gameplay Stats')
            stats.addline(' %s / %s Ts (%s%s)' % (tWins, tTotal, t_win_ratio, '%'))
            stats.addline(' %s / %s CTs (%s%s)' % (ctWins, ctTotal, ct_win_ratio, '%'))
            stats.addline(' Total: %s / %s (%s%s)' % (tWins + ctWins, tTotal + ctTotal, convertKdr(tWins + ctWins, tTotal + ctTotal, True), '%'))
            
            stats.send(self.userid)
            
    def showNext(self):
        if self.exists():
            if database.rank(self.steam) != None:
                if int(self.position) == 1: return es.tell(self.userid, '#green', 'There are no players ahead of you.')
                position = database.getPosition(self.steam)
                allplayers = database.getAll()
                del allplayers[:position-10]
                del allplayers[position:]
                if allplayers:
                    i = 1
                    for mypoints in database.getPoints(self.steam):
                        next = popuplib.create('next_%s' % self.userid)
                        next.addline('->1. Next Players')
                        next.addline(' ')
                        for steamid, kills, deaths, points in allplayers:
                            if steamid.startswith('BOT_'):
                                name = steamid.replace('BOT_', '')
                            else:
                                name = database.getPlayerName(steamid)
                        
                            difference = mypoints - points
                        
                            if i < 11:
                                if i == 10:
                                    mykills, mydeaths, mypoints = database.rank(self.steam)
                                    kdr = convertKdr(mykills, mydeaths)
                                    next.addline(' ')
                                    next.addline('  %s  %s  %s (KDR %s)' % (database.getPosition(steamid) if not database.getPosition(steamid) < 10 else '0' + database.getPosition(steamid), getFormat(mypoints), name, kdr))
                                else:
                                    kdr = convertKdr(kills, deaths)
                                    next.addline('  %s  +%s  %s (KDR %s)' % (database.getPosition(steamid) if not database.getPosition(steamid) < 10 else '0' + database.getPosition(steamid), abs(difference), name, kdr))
                            
                            i += 1
                            
                        next.addline(' ')
                        next.addline('0. Close')
                        next.send(self.userid)
                        
    def showHitbox(self):
        if self.exists():
            head, chest, stomach, leftarm, rightarm, leftleg, rightleg = database.getHitboxData(self.steam)
            damage = database.getDamage(self.steam)
            total = head+chest+stomach+leftarm+rightarm+leftleg+rightleg
            
            hitbox = popuplib.create('hitbox_%s' % self.userid)
            hitbox.addline('%s - Hitbox' % es.getplayername(self.userid))
            hitbox.addline(' ')
            hitbox.addline('->1. Statistics')
            hitbox.addline(' %s Damage' % damage)
            hitbox.addline(' %s Hits' % (head+chest+stomach+leftarm+rightarm+leftleg+rightleg))
            hitbox.addline(' ')
            hitbox.addline('->2. Head (%s)' % head)
            hitbox.addline(' %s (%s%s)' % (makeBar(head, total), convertKdr(head, total, True), '%'))
            hitbox.addline(' ')
            hitbox.addline('->3. Chest (%s)' % chest)
            hitbox.addline(' %s (%s%s)' % (makeBar(chest, total), convertKdr(chest, total, True), '%'))
            hitbox.addline(' ')
            hitbox.addline('->4. Stomach (%s)' % stomach)
            hitbox.addline(' %s (%s%s)' % (makeBar(stomach, total), convertKdr(stomach, total, True), '%'))
            hitbox.addline(' ')
            hitbox.addline('->5. Left Arm (%s)' % leftarm)
            hitbox.addline(' %s (%s%s)' % (makeBar(leftarm, total), convertKdr(leftarm, total, True), '%'))
            hitbox.addline(' ')
            hitbox.addline('->9. Next')
            hitbox.addline('0. Close')
            
            hitbox.send(self.userid)
            hitbox.menuselect = self.secondHitbox
            
    def secondHitbox(self, userid, choice, popupid):
        if not self.exists(): return
        if int(choice) == 9:
            head, chest, stomach, leftarm, rightarm, leftleg, rightleg = database.getHitboxData(self.steam)
            damage = database.getDamage(self.steam)
            total = head+chest+stomach+leftarm+rightarm+leftleg+rightleg
            
            hitbox2 = popuplib.create('hitbox2_%s' % self.userid)
            hitbox2.addline('%s - Hitbox' % es.getplayername(self.userid))
            hitbox2.addline(' ')
            hitbox2.addline('->1. Statistics')
            hitbox2.addline(' %s Damage' % damage)
            hitbox2.addline(' %s Hits' % (head+chest+stomach+leftarm+rightarm+leftleg+rightleg))
            hitbox2.addline(' ')
            hitbox2.addline('->6. Right Arm (%s)' % rightarm)
            hitbox2.addline(' %s (%s%s)' % (makeBar(rightarm, total), convertKdr(rightarm, total, True), '%'))
            hitbox2.addline(' ')
            hitbox2.addline('->7. Left Leg (%s)' % leftleg)
            hitbox2.addline(' %s (%s%s)' % (makeBar(leftleg, total), convertKdr(leftleg, total, True), '%'))
            hitbox2.addline(' ')
            hitbox2.addline('->8. Right Leg (%s)' % rightleg)
            hitbox2.addline(' %s (%s%s)' % (makeBar(rightleg, total), convertKdr(rightleg, total, True), '%'))
            hitbox2.addline(' ')
            hitbox2.addline('->9. Back')
            hitbox2.addline('0. Close')
            
            hitbox2.send(self.userid)
            hitbox2.menuselect = self.backHitbox
            
        else:
            popuplib.delete('hitbox_%s' % self.userid)
            
    def backHitbox(self, userid, choice, popupid):
        if int(choice) == 9:
            popuplib.send('hitbox_%s' % self.userid , self.userid)
        else:
            for x in [1, 2]:
                popuplib.delete('hitbox%s_%s' % (x, self.userid))
                
    def showWeapons(self):
        if self.exists():
            weapons = database.getWeaponData(self.steam)
            self.makeWeaponPage(weapons)
            popuplib.send('%s_weapons_1' % self.userid, self.userid)
            
    def makeWeaponPage(self, weapons):
        v = 0
        i = 0
        p = popuplib.create('%s_weapons_1' % self.userid)
        kills = database.getKills(getPlayer(self.userid).getUniqueId())
            
        p.addline('%s - Weapons (1/4)' % (es.getplayername(self.userid)))
        p.addline(' ')
        p.addline('%s Total Kills' % kills)
        p.addline(' ')
        weapons_new = []
        for weapon in weapons:
            weapons_new.append(weapon)
        del weapons_new[:1]
        weapons = weapons_new
        for weapon in weapons[:7]:
            i += 1
            v += 1
            p.addline('->%s. %s (%s)' % (v, all_weapons[i-1].capitalize(), weapon))
            p.addline(' %s (%s%s)' % (makeBar(weapon, kills), convertKdr(weapon, kills, True), '%'))
            p.addline(' ')
            
        p.addline('->9. Next')
        p.addline('0. Close')
        p.menuselect = self.weaponsSelect
            
        v = 0
        p2 = popuplib.create('%s_weapons_2' % self.userid)
        p2.addline('%s - Weapons (2/4)' % (es.getplayername(self.userid)))
        p2.addline(' ')
        for weapon in weapons[7:14]:
            i += 1
            v += 1
            p2.addline('->%s. %s (%s)' % (v, all_weapons[i-1].capitalize(), weapon))
            p2.addline(' %s (%s%s)' % (makeBar(weapon, kills), convertKdr(weapon, kills, True), '%'))
            p2.addline(' ')
            
        p2.addline('->8. Back')
        p2.addline('->9. Next')
        p2.addline('0. Close')
        p2.menuselect = self.weaponsSelect
            
        v = 0
        p3 = popuplib.create('%s_weapons_3' % self.userid)
        p3.addline('%s - Weapons (3/4)' % (es.getplayername(self.userid)))
        p3.addline(' ')
        p3.addline('%s Total Kills' % kills)
        p3.addline(' ')
        for weapon in weapons[14:21]:
            i += 1
            v += 1
            p3.addline('->%s. %s (%s)' % (v, all_weapons[i-1].capitalize(), weapon))
            p3.addline(' %s (%s%s)' % (makeBar(weapon, kills), convertKdr(weapon, kills, True), '%'))
            p3.addline(' ')
            
        p3.addline('->8. Back')
        p3.addline('->9. Next')
        p3.addline('0. Close')
        p3.menuselect = self.weaponsSelect
            
        v = 0
        p4 = popuplib.create('%s_weapons_4' % self.userid)
        p4.addline('%s - Weapons (4/4)' % (es.getplayername(self.userid)))
        p4.addline(' ')
        p4.addline('%s Total Kills' % kills)
        p4.addline(' ')
        for weapon in weapons[21:28]:
            i += 1
            v += 1
            p4.addline('->%s. %s (%s)' % (v, all_weapons[i-1].capitalize(), weapon))
            p4.addline(' %s (%s%s)' % (makeBar(weapon, kills), convertKdr(weapon, kills, True), '%'))
            p4.addline(' ')
            
        p4.addline('->8. Back')
        p4.addline('0. Close')
        p4.menuselect = self.weaponsSelect
                        
    def weaponsSelect(self, userid, choice, popupid):
        number = int(popupid.split('_')[2])
        if number == 1:
            if choice == 9:
                popuplib.send('%s_weapons_2' % self.userid, self.userid)
                
        elif number == 2:
            if choice == 8:
                popuplib.send('%s_weapons_1' % self.userid, self.userid)
            elif choice == 9:
                popuplib.send('%s_weapons_3' % self.userid, self.userid)
                
        elif number == 3:
            if choice == 8:
                popuplib.send('%s_weapons_2' % self.userid, self.userid)
            elif choice == 9:
                popuplib.send('%s_weapons_4' % self.userid, self.userid)
                
        elif number == 4:
            if choice == 8:
                popuplib.send('%s_weapons_3' % self.userid, self.userid)

################################################################
##
## User Setting Class

class UserSetting(object):
    saycommand = None
    
    def __init__(self, setting, initial):
        self.players = {}
        self.setting = setting
        self.initial = initial
        self.callbacks = {}
        
    def registerPlayer(self, userid):
        if int(userid) not in self.players:
            self.players[int(userid)] = self.initial
            
    def change(self, userid, value):
        if int(userid) not in self.players:
            self.registerPlayer(int(userid))
            
        self.players[int(userid)] = value
        
    def get(self, userid):
        if int(userid) not in self.players:
            self.registerPlayer(int(userid))
            
        return self.players[int(userid)]
        
    def register(self, command, callback):
        cmdlib.registerSayCommand(command, self.callback, 'Ranks - %s command' % self.setting)
        self.function = callback
        self.saycommand = command
        
    def callback(self, userid, args):
        function = self.function
        self.function(userid, args, self.get(userid))
        
    def unregister(self):
        cmdlib.unregisterSayCommand(self.saycommand)
        self.saycommand = None

usersettings = {}
def addUserSetting(usersetting, command, callback, initial):
    if usersetting in usersettings:
        return es.dbgmsg(0, 'The setting %s is already registered' % usersetting)
        
    usersettings[usersetting] = UserSetting(usersetting, initial)
    if command: usersettings[usersetting].register(command, callback)
    return usersettings[usersetting]
        
    
################################################################
##
## Announcer Class
    
class Announcer(object):
    def __init__(self, name, freq, message):
        self.name = name
        self.freq = freq
        self.message = message
        self.round = 0
            
    def count(self):
        self.round += 1
        self.check()
        
    def check(self):
        if self.round == self.freq:
            for i in es.getUseridList():
                saytext2(i, playerlib.getPlayer(i).attributes['index'], self.message)
                
            self.round = 0
            
announcements = {}
def addAnnounce(name, freq, message=None):
    if name in announcements:
        return es.dbgmsg(0, 'The announcement %s is already registered' % name)
        
    announcements[name] = Announcer(name, freq, message)
    return announcements[name]
        

################################################################
##
## Saytext2 Wrapper

colors = ('#blue','#grey','#red','#team')
teams = {'#red':'#t','#blue':'#ct'}

def saytext2(users, index, message):
    index, team = getIndex(index,message)
    message = message.replace('#default','\x01').replace('#green', '\x04').replace('#lightgreen','\x03')
    for color in colors:
        message = message.replace(color, '\x03')
    usermsg.saytext2(users,index,message)

def getIndex(index, message):
    if '#team' in message:
        return (index,es.getindexprop(index,'CBaseEntity.m_iTeamNum'))
    for color in teams:
        if color in message:
            players = playerlib.getPlayerList(teams[color])
            if players:
                return (players[0].index,players[0].team)
    if '#grey' in message:
        return (-1,0)
    return (0,0)


######################################

''' Load / Unload Functions '''
    
def load():
    es.dbgmsg(0, '[Crazy Ranks] Beginning main script load...')
    es.dbgmsg(0, '[Crazy Ranks] Executed main configuration file')
    start()
    es.dbgmsg(0, '[Crazy Ranks] Executed the main script. Script is ready.')
    database.createNameDatabase()
    for i in es.getUseridList():
        database.setPlayerName(getPlayer(i).getUniqueId(), es.getplayername(i))
        
    cmdlib.registerServerCommand('crazyranks', clientCommand, 'Crazyranks Main Command')
    
def unload():
    database.close()
    unloadModules()
    for setting in settings:
        settings[setting].unregister()
        
    cmdlib.unregisterServerCommand('crazyranks')
        
def server_shutdown(ev):
    database.close()
    
######################################

players = {}

def getPlayer(userid):
    if userid not in players:
        players[userid] = Player(userid)
        return players[userid]
        
    return players[userid]
    
def es_map_start(ev):
    # Save the changes on map start
    database.commit()
    for announce in announcements:
        announcements[announce].round = 0
    
######################################
    
module_file = es.getAddonPath("crazyranking") + '/config/addons.txt'
modules_to_import = []
loaded_modules = []

def start():
    file = open(module_file, 'r')
    lines = file.readlines()
    for line in filter(lambda line: not line.startswith('//') and not line.startswith('\n'), lines):
        modules_to_import.append(line.replace('load_module', '').strip())
            
    loadModules()
    
def loadModules():
    for module in modules_to_import:
        p = es.getAddonPath("crazyranking") + '/modules/%s/%s.py' % (module, module)
        if path(p).isfile():
            es.load('crazyranking/modules/%s' % module)
            loaded_modules.append(module)
        
def unloadModules():
    for module in loaded_modules:
        es.unload('crazyranking/modules/%s' % module)
        loaded_modules.remove(module)
    
''' Event Functions '''

def player_activate(ev):
    # Player Activate (event)
    
    # Adds the player to the names database if he isn't
    # already registered.
    userid = int(ev['userid'])
    steamid = getPlayer(userid).getUniqueId()
    database.setPlayerName(steamid, es.getplayername(userid))
    createConsoleOutput(userid)
    if steamid not in database:
        database.addDefaultPoints(steamid)
    
def player_changename(ev):
    userid = int(ev['userid'])
    name = ev['newname']
    steamid = getPlayer(userid).getUniqueId()
    database.setPlayerName(steamid, name)

def player_team(ev):
    # Player Team (event)
    
    # Adds the player to a global database so his total gameplay time 
    # can be calculated later on.
    userid = int(ev['userid'])
    steamid = getPlayer(userid).getUniqueId()
    if userid not in players_conn:
        
        players_conn[userid] = {}
        players_conn[userid]['steam'] = steamid
        players_conn[userid]['time'] = time.time()
        
        if settings.message_player_connect_event:
            points = database.getPoints(steamid)
            if points:
                total = database.getTotalCount()
                position = database.getPosition(steamid)
                index = playerlib.getPlayer(userid).attributes['index']
                msg(index, 'player connect', {'player': es.getplayername(userid), 'points' : getFormat(points), 'position' : position, 'total' : total})
                    
    for usersetting in usersettings:
        usersettings[usersetting].registerPlayer(userid)

def player_disconnect(ev):
    # Player Disconnect (event)
    
    # Checks whether the player disconnected so it adds
    # time played to his total game time on the server.
    userid = int(ev['userid'])
    if userid in players_conn:
        steamid = players_conn[userid]['steam']
        timeplayed = time.time() - players_conn[userid]['time']
        
        database.addTime(steamid, timeplayed)
        del players_conn[userid]
    
def player_death(ev):
    # Player Death (event)
    
    # Passes the info the database, including the victim,
    # attacker and whether the kill was a headshot.
    userid=None
    hs=False
    weapon = ev['weapon'].replace('weapon_', '')
    vic = getPlayer(int(ev['userid']))
    att = getPlayer(int(ev['attacker']))
    victim = vic.getUniqueId()
    attacker = att.getUniqueId()
    if int(ev['headshot']): hs=True
    if settings.message_player_events: userid = int(ev['userid'])
    database.death(victim, attacker, weapon, hs, userid, int(ev['attacker']))
    
def player_hurt(ev):
    # Player Hurt (event)
    
    # Passes the info to the database, including the attacker,
    # the hitbox and the damage.
    player = getPlayer(ev['attacker'])
    attacker = player.getUniqueId()
    hit = int(ev['hitgroup'])
    damage = int(ev['dmg_health'])
    index = playerlib.getPlayer(ev['userid']).attributes['index']
    database.hurt(attacker, damage, int(ev['userid']))
    database.hitbox(attacker, hit)
    
def round_start(ev):
    for announce in announcements:
        announcements[announce].count()
        
def round_end(ev):
    # Round End (event)
    
    # Pass the winner to the database so he can get the bonus.
    winner = int(ev['winner'])
    database.roundend(winner)
       
    random_winner = playerlib.getPlayerList('#ct' if winner == 3 else '#t')[0]
    index = playerlib.getPlayer(random_winner).attributes['index']
    if index: msg(index, 'round end', {'team' : 'Terrorists' if winner==2 else 'Counter-Terrorists', 'bonus' : settings.round_end_winner_bonus})
    
def hostage_killed(ev):
    # Hostage Killed (event)
    
    # Passes the user that killed the hostage to the database.
    steamid = getPlayer(int(ev['userid'])).getUniqueId()
    database.hostage(steamid, 0, int(ev['userid']))
    
def hostage_rescued(ev):
    # Hostage Rescue (event)
    
    # Passes the user that saved the hostage to the database.
    steamid = getPlayer(int(ev['userid'])).getUniqueId()
    database.hostage(steamid, 1, int(ev['userid']))
    
def bomb_planted(ev):
    # Bomb Plant (event)
    
    # Passes the user that planted the bomb to the database.
    steamid = getPlayer(int(ev['userid'])).getUniqueId()
    database.bomb(steamid, 0, int(ev['userid']))
   
def bomb_defused(ev):
    # Bomb Defuse (event)
    
    # Passes the user that defused the bomb to the database.
    steamid = getPlayer(int(ev['userid'])).getUniqueId()
    database.bomb(steamid, 1, int(ev['userid']))
   
def bomb_exploded(ev):
    # Bomb Explode (event)
    
    # Passes the user that planted the bomb which exploded to the database.
    steamid = getPlayer(int(ev['userid'])).getUniqueId()
    database.bomb(steamid, 2, int(ev['userid']))
   
def player_say(ev):
    # Player Say (chat) (event)
    
    # Checks what the player says so the correct menu shows up.
    
    userid = int(ev['userid'])
    args = ev['text']
    
    if ev['text'] == settings.rank_command:
        rankCommand(userid, args)
       
    if ev['text'] == settings.stats_command:
        statsCommand(userid, args)
    
    if ev['text'] == settings.hitbox_command:
        hitboxCommand(userid, args)
        
    if ev['text'] == settings.next_command:
        nextCommand(userid, args)
       
    if ev['text'].startswith('top'):
        topCommand(userid, args)
        
    if ev['text'] == settings.weapon_command:
        weaponCommand(userid, args)
        
   
''' Custom Functions '''

def convertKdr(kills, deaths, percent_string=False):
    # Converts the kdr to something the player can read and rounds is to 2 decimals
    # Used to get the percent of others things aswell.
    if percent_string:
        if kills == 0 or deaths == 0:
            return '0'
        
    kills = int(kills)
    deaths = int(deaths)
    kills_kdr = None
    deaths_kdr = None
    
    if kills == 0:
        kills_kdr = 1
    
    if deaths == 0:
        deaths_kdr = 1
    
    if kills_kdr or deaths_kdr:
        if kills_kdr and not deaths_kdr:
            kdr = '%.2f' % float(kills_kdr/deaths)
        
        if deaths_kdr and not kills_kdr:
            kdr = '%.2f' % float(kills/deaths_kdr)
        
        if deaths_kdr and kills_kdr:
            kdr = '0.00'
            
        return kdr
    
    else:
        if percent_string:
            kdr = '%i' % int((kills/deaths)*100)
        else:
            kdr = '%.2f' % float(kills/deaths)
             
        return kdr
   
def getWinRatio(t_rounds, t_total, ct_rounds, ct_total):
    # Used to get the user's win ratio on a team
    # Rounds it to a full number in percents
    
    if t_rounds == 0: t_rounds = 1
    if t_total == 0: t_total = 1
    if ct_rounds == 0: ct_rounds = 1
    if ct_total == 0: ct_total = 1
    
    ct_wins = float(ct_rounds/ct_total)*100
    t_wins = float(t_rounds/t_total)*100
    
    ct_win_ratio = '%i' % int(ct_wins)
    t_win_ratio = '%i' % int(t_wins)
    
    return t_win_ratio, ct_win_ratio

def makeBar(value, count):
    # Makes the ||| lines which are seen in 'hitbox' popup
    # If a server owner thinks that the bars are too big, he may
    # edit the '42' below with another value (lower in this case)
    #es.msg('%s - %s' % (value, count))
    ratio = int((int(value)/int(count))*42)
    return '|' * ratio
        
def getFormat(value):
    value = str(value)
    if len(value) >= 3:
        value = value[0:len(value) - 3] + ',' + value[len(value) - 3: len(value)]
       
    return value

def GetInHMS(seconds):
    seconds = int(seconds)
    hours = int(seconds / 3600)
    seconds -= 3600*hours
    minutes = int(seconds / 60)
    seconds -= 60*minutes
    return "%02dh %02dm %02ds" % (int(hours), int(minutes), (seconds))
    
def createConsoleOutput(userid):
    lines = [
        '*************************************************',
        '----- Welcome %s!' % es.getplayername(userid),
        'This server is running the Crazy Ranks plugin.',
        'Available commands:',
        ' ',
        '%s - Shows your statistics to other players' % settings.rank_command,
        '%s - Shows detailed statistics about yourself' % settings.stats_command,
        '%s - Shows the players infront of you' % settings.next_command,
        '%s - Shows detailed statistics of your hitbox' % settings.hitbox_command,
        '%s - Shows detailed statistics about your weapons' % settings.weapon_command,
        '%s - Shows the top players on the server' % 'top',
        ' ',
        ' ',
        'Developed by .craziesT ',
        'Forum link for official support and requests:',
        '> http://forum.crazie.org',
        ' ',
        '*************************************************'
        ]
    for line in lines:
        usermsg.echo(userid, line)
        
    es.tell(userid, '#green', 'Open the console for information about the ranks plugin.')
    
#############################################################################

def clientCommand(args):
    if args[0] == 'export':
        export.export()

def rankCommand(userid, args):
    # The rank command
    # Gets the players kills, deaths, kdr, position and displays it in
    # a nice looking chat line which can be seen by other players.
    
    player = getPlayer(userid)
    player.showRank()
    
def topCommand(userid, args):
    # The top command
    # Displays the top10 players in the database
    top10 = False
    
    # TODO: topx (x = 10, 20...)
    # Need to test if it works...
    if database.getTotalCount() < 10:
        es.tell(userid, '#green', 'Not enough players are ranked yet.')
        return
    
    if args == 'top': top10 = True
    
    i = 1
    if settings.points_system:
        if top10:
            topCommand(userid, args='top10')
                    
        else:
            positionsearch = args.replace('top', '').strip()
            positionsearch = int(positionsearch)
            allplayers = database.getAll()
            if len(allplayers) < int(positionsearch):
                positionsearch = len(allplayers)
            del allplayers[:int(positionsearch) - 10]
            del allplayers[int(positionsearch):]
            
            if allplayers:
                i = 1
                top = popuplib.create('top_%s_%s' % (userid, int(positionsearch)))
                if int(settings.points_system):
                    top.addline('->1. Top (%s-%s) Players by Points' % (int(positionsearch) - 9 if int(positionsearch) > 9 else 1, int(positionsearch)))
                else:
                    top.addline('->1. Top (%s-%s) Players by Kills' % (int(positionsearch) + 1, int(positionsearch)))
                top.addline(' ')
                top.menuselect = topHandler
                for steamid, kills, deaths, points in allplayers:
                    if steamid.startswith('BOT_'):
                        if settings.do_bots_count:
                            if settings.show_bots_on_top:
                                name = steamid.replace('BOT_', '')
                            else: continue
                        else: continue
                    else:
                        name = database.getPlayerName(steamid)
                
                    kdr = convertKdr(kills, deaths)
                    if i < 11:
                        if int(settings.points_system):
                            if int(positionsearch) < 11:
                                if i < 10:
                                    #es.msg(str(positionsearch) + ',' + str(points) + ',' + str(name) + ',' + str(kdr))
                                    top.addline('  0' + str(int(positionsearch) - 10+i) + '  %s  %s (KDR %s)' % (str(points), str(name), str(kdr)))
                                else:
                                    top.addline('  ' + str(int(positionsearch) - 10+i) + '  %s  %s (KDR %s)' % (str(points), str(name), str(kdr)))
                            else:
                                top.addline('  ' + str(int(positionsearch) - 10+i) + '  %s  %s (KDR %s)' % (str(points), str(name), str(kdr)))
                        else:
                            if int(positionsearch) < 11:
                                if i < 10:
                                    top.addline('  0' + str(int(positionsearch) - 10+i) + '  %s  %s (KDR %s)' % (str(points), str(name), str(kdr)))
                                else:
                                    top.addline('  ' + str(int(positionsearch) - 10+i) + '  %s  %s (KDR %s)' % (str(points), str(name), str(kdr)))
                            else:
                                top.addline('  ' + str(int(positionsearch) - 10+i) + '  %s  %s (KDR %s)' % (str(points), str(name), str(kdr)))
                            
                        i += 1
                        
                    else:
                        break
                    
                top.addline(' ')
                if not int(positionsearch) < 11:
                    top.addline('->8. Back')
                    
                if not database.getTotalCount() < int(positionsearch):
                    top.addline('->9. Next')
                top.addline('0. Close')
                    
                top.send(userid)
    
def topHandler(userid, choice, popupid):
    last_split = popupid.split('_')
    last_search = int(last_split[2])
    if choice == 8:
        if last_search < 0:
            topCommand(userid, "top10")
        
        elif last_search < 11:
            popuplib.send(popupid, userid)
        else:
            newsearch = int(last_search)-10
            if newsearch < 11:
                topCommand(userid, "top10")
                return
            topCommand(userid, "top%i" % newsearch)
            
    elif choice == 9:
        if last_search >= database.getTotalCount():
            popuplib.send(popupid, userid)
        else:
            newsearch = int(last_search)+10
            if newsearch >= database.getTotalCount():
                topCommand(userid, "top%i" % database.getTotalCount())
                return
            topCommand(userid, "top%i" % newsearch)
        

def statsCommand(userid, args):
    # The 'statsme' command
    # Displays all of the players available except the hitbox and weapon
    # statistics which can be viewed specifically
    
    player = getPlayer(userid)
    player.showStats()
    
def nextCommand(userid, args):
    player = getPlayer(userid)
    player.showNext()
    
def hitboxCommand(userid, args):
    # The 'hitboxme' command
    # Shows the players hitbox data including total damage, total hits on a body part
    # and the percentage of hits on that part
    
    player = getPlayer(userid)
    player.showHitbox()
    
def weaponCommand(userid, args):
    player = getPlayer(userid)
    player.showWeapons()
    
import export
    